from __future__ import annotations

from typing import Generic, Optional, Sequence, TypeVar

import numpy as np
from jax.tree_util import tree_map
from matplotlib.axes import Axes

from .annotations import PyTree, RealArray, SliceLike
from .dataclasses import dataclass
from .leaky_integral import leaky_integrate_time_series

__all__ = ['PlottableTrajectory']


Trajectory = TypeVar('Trajectory', bound=PyTree)


@dataclass
class PlottableTrajectory(Generic[Trajectory]):
    trajectory: Trajectory
    "The trajectory is a PyTree containing the plotted data in its nonstatic attributes."
    times: np.ndarray
    "The times corresponding to the data points in each of the plotted attributes."

    def plot(self,
             data: RealArray,
             axis: Axes,
             *,
             title: Optional[str] = None,
             legend: int = 0,
             labels: Optional[Sequence[str]] = None,
             simple_label: Optional[str] = None,
             label_function: Optional[Callable[[Tuple[int, ...]], str]] = None,
             decay: Optional[float] = None,
             clip_slice: SliceLike = slice(None),
             clip_boolean: Optional[RealArray] = None) -> None:
        """
        Plot the PlottableTrajectory into a matplotlib axis.
        Args:
            data: The data to be plotted.  Typically, this is a sub-element of trajectory.
            axis: A matplotlib axis to plot into.
            title: The title of the subplot.
            legend: The maximum number of entries for which to generate a legend.
            labels: The labels corresponding to each plot.
            simple_label: A single label with which to construct the labels.
            label_function: A function with which to construct the labels.
            decay: If not none, applies a leaky-integral with the given decay to the plotted points.
            clip_slice: Restrict the plot to a slice.
            clip_boolean: Restrict the plot to elements where this array is true.
        """
        if title is not None:
            axis.set_title(title)

        unclipped_times = self.times[clip_slice]
        all_ys = np.asarray(data)
        if not np.all(np.isfinite(all_ys)):
            print(f"Graph {'' if title is None else title + ' '}contains infinite numbers.")
            all_ys = np.where(np.isfinite(all_ys), all_ys, 0.0)

        plot_indices = list(np.ndindex(all_ys.shape[1:]))
        if labels is not None:
            if len(plot_indices) != len(labels):
                raise ValueError
        elif simple_label is not None:
            if not plot_indices:
                labels = []
            elif len(plot_indices) == 1:
                labels = [simple_label]
            elif all_ys.ndim == 2:
                labels = [f"{simple_label} {x[0]}" for x in plot_indices]
            else:
                labels = [f"{simple_label} {x}" for x in plot_indices]
        elif label_function is not None:
            labels = [label_function(x) for x in plot_indices]
        else:
            labels = [str(x) for x in plot_indices]

        for plot_index, label in zip(plot_indices, labels):
            ys = all_ys[(clip_slice, *plot_index)]
            if decay is not None:
                ys = leaky_integrate_time_series(ys, decay)
            if clip_boolean is not None:
                ys = np.array(ys)
                ys[~clip_boolean] = np.nan
                times = np.array(unclipped_times)
                times[~clip_boolean] = np.nan
            else:
                times = unclipped_times
            axis.plot(times, ys, label=label)
        number_of_graph_lines = np.prod(data.shape[1:])
        if legend > 0 and 0 < number_of_graph_lines <= legend:
            axis.legend()

    def slice_into(self, s: SliceLike) -> Trajectory:
        "Return a new trajectory whose nonstatic attributed are sliced."
        return tree_map(lambda x: x[s], self.trajectory)



from ._src.dataclasses.dataclass import MISSING, FrozenInstanceError, InitVar, dataclass
from ._src.dataclasses.helpers import (Field, asdict, astuple, document_dataclass, field,
                                       field_names, field_names_and_values,
                                       field_names_values_metadata, field_values, fields,
                                       is_dataclass, replace)

__all__ = ['Field', 'FrozenInstanceError', 'InitVar', 'MISSING', 'asdict', 'astuple', 'dataclass',
           'document_dataclass', 'field', 'field_names', 'field_names_and_values',
           'field_names_values_metadata', 'field_values', 'fields', 'is_dataclass', 'replace']



import dataclasses
from dataclasses import MISSING, Field, asdict, astuple
from dataclasses import fields as d_fields
from dataclasses import is_dataclass, replace
from typing import (Any, Callable, Iterable, Mapping, MutableMapping, Optional, Tuple, TypeVar,
                    overload)

__all__ = ['field', 'Field', 'fields', 'asdict', 'astuple', 'replace', 'is_dataclass',
           'field_names', 'field_names_and_values', 'field_names_values_metadata', 'field_values',
           'document_dataclass']


T = TypeVar('T', bound=Any)


# NOTE: Actual return type is 'Field[T]', but we want to help type checkers
# to understand the magic that happens at runtime.
# pylint: disable=redefined-builtin
@overload  # `default` and `default_factory` are optional and mutually exclusive.
def field(*, static: bool = False, default: T, init: bool = ..., repr: bool = ...,
          hash: Optional[bool] = ..., compare: bool = ...,
          metadata: Optional[Mapping[str, Any]] = ...) -> T:
    ...


@overload
def field(*, static: bool = False, default_factory: Callable[[], T], init: bool = ...,
          repr: bool = ..., hash: Optional[bool] = ..., compare: bool = ...,
          metadata: Optional[Mapping[str, Any]] = ...) -> T:
    ...


@overload
def field(*, static: bool = False, init: bool = ..., repr: bool = ..., hash: Optional[bool] = ...,
          compare: bool = ..., metadata: Optional[Mapping[str, Any]] = ...) -> Any:
    ...


def field(*, static: bool = False, default: Any = MISSING,
          default_factory: Callable[[], Any] = MISSING, init: bool = True,  # type: ignore
          repr: bool = True, hash: Optional[bool] = None, compare: bool = True,
          metadata: Optional[Mapping[str, Any]] = None) -> Any:
    """
    Args:
        static: Indicates whether a field is a pytree or static.  Pytree fields are
            differentiated and traced.  Static fields are hashed and compared.
    """
    if metadata is None:
        metadata = {}
    return dataclasses.field(metadata={**metadata, 'static': static},
                             default=default, default_factory=default_factory, init=init, repr=repr,
                             hash=hash, compare=compare)  # type: ignore


def fields(d: Any, *, static: Optional[bool] = None) -> Iterable[Field[Any]]:
    if static is None:
        yield from d_fields(d)
    for this_field in d_fields(d):
        if this_field.metadata.get('static', False) == static:
            yield this_field


def field_names(d: Any, *, static: Optional[bool] = None) -> Iterable[str]:
    for this_field in fields(d, static=static):
        yield this_field.name


def field_names_and_values(d: Any, *, static: Optional[bool] = None) -> Iterable[Tuple[str, Any]]:
    for name in field_names(d, static=static):
        yield name, getattr(d, name)


def field_values(d: Any, *, static: Optional[bool] = None) -> Iterable[Any]:
    for name in field_names(d, static=static):
        yield getattr(d, name)


def field_names_values_metadata(d: Any, *, static: Optional[bool] = None) -> (
        Iterable[Tuple[str, Any, Mapping[str, Any]]]):
    for this_field in fields(d, static=static):
        yield this_field.name, getattr(d, this_field.name), this_field.metadata


def document_dataclass(pdoc: MutableMapping[str, Any], name: str) -> None:
    pdoc[f'{name}.static_fields'] = False
    pdoc[f'{name}.nonstatic_fields'] = False
    pdoc[f'{name}.tree_flatten'] = False
    pdoc[f'{name}.tree_unflatten'] = False
    pdoc[f'{name}.display'] = False
    pdoc[f'{name}.replace'] = False



from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Mapping, Optional, Sequence, Tuple, TypeVar, Union

import jax.numpy as jnp
from jax import vjp
from jax.experimental.host_callback import id_print
from jax.tree_util import tree_map

from .shims import custom_vjp

__all__ = ['copy_cotangent', 'block_cotangent', 'replace_cotangent', 'print_cotangent',
           'CotangentMapper']


X = TypeVar('X')


# copy_cotangent -----------------------------------------------------------------------------------
def copy_cotangent(x: X, y: X) -> Tuple[X, X]:
    return x, y


# Apply after to work around mypy deficiency.
copy_cotangent = custom_vjp(copy_cotangent)  # type: ignore


def _copy_cotangent_fwd(x: X, y: X) -> Tuple[Tuple[X, X], None]:
    return (x, y), None


def _copy_cotangent_bwd(residuals: None, xy_bar: Tuple[X, X]) -> Tuple[X, X]:
    del residuals
    x_bar, _ = xy_bar
    return x_bar, x_bar


copy_cotangent.defvjp(_copy_cotangent_fwd, _copy_cotangent_bwd)  # type: ignore


# replace_cotangent --------------------------------------------------------------------------------
def replace_cotangent(x: X, new_cotangent: X) -> X:
    return x


# Apply after to work around mypy deficiency.
replace_cotangent = custom_vjp(replace_cotangent)  # type: ignore


def _replace_cotangent_fwd(x: X, new_cotangent: X) -> Tuple[X, X]:
    return x, new_cotangent


def _replace_cotangent_bwd(residuals: X, x_bar: X) -> Tuple[X, X]:
    return residuals, x_bar


replace_cotangent.defvjp(_replace_cotangent_fwd, _replace_cotangent_bwd)  # type: ignore


# block_cotangent ----------------------------------------------------------------------------------
def block_cotangent(f: Callable[..., X],
                    block_argnums: Union[int, Tuple[int, ...]],
                    static_argnums: Union[int, Tuple[int, ...]] = ()) -> Callable[..., X]:
    if isinstance(block_argnums, int):
        block_argnums = (block_argnums,)
    set_block_argnums = set(block_argnums)

    @partial(custom_vjp, static_argnums=static_argnums)
    def blocked_f(*args: Any, **kwargs: Any) -> Any:
        return f(*args, **kwargs)

    def blocked_f_fwd(*args: Any, **kwargs: Any) -> Tuple[Any, Any]:
        return vjp(f, *args, **kwargs)  # type: ignore

    def blocked_f_bwd(residuals: Any, output_bar: Any) -> Tuple[Any, ...]:
        f_vjp = residuals
        input_bar = f_vjp(output_bar)
        return tuple(None if i in set_block_argnums else x_bar
                     for i, x_bar in enumerate(input_bar))

    blocked_f.defvjp(blocked_f_fwd, blocked_f_bwd)

    return blocked_f  # type: ignore


# print_cotangent ----------------------------------------------------------------------------------
def print_cotangent(x: X, what: Optional[str] = None) -> X:
    return x


def _print_cotangent_fwd(x: X, what: Optional[str]) -> Tuple[X, None]:
    return x, None


def _print_cotangent_bwd(what: Optional[str], residuals: None, x_bar: X) -> Tuple[X]:
    del residuals
    x_bar = (id_print(x_bar)  # type: ignore
             if what is None
             else id_print(x_bar, what=what))  # type: ignore
    return (x_bar,)


# Apply after to work around mypy deficiency.
print_cotangent = custom_vjp(print_cotangent, static_argnums=1)  # type: ignore


print_cotangent.defvjp(_print_cotangent_fwd, _print_cotangent_bwd)  # type: ignore


# print_cotangent ----------------------------------------------------------------------------------
_T = TypeVar('_T')


@dataclass
class CotangentMapper:
    cotangent_names: Sequence[str]

    def to_mapping(self, cotangents: _T) -> Mapping[str, _T]:
        """
        Args:
            cotangents: A PyTree whose nonstatic fields are arrays representing the group of
                cotangents.
        Returns: A mapping from cotangent names to sub-arrays representing individual cotangents.
        """
        return {cotangent_name: tree_map(lambda x: x[index], cotangents)
                for index, cotangent_name in enumerate(self.cotangent_names)}

    def vmap(self, x: _T, index: Optional[int]) -> _T:
        """
        Args:
            x: The value of the nonzero cotangent.
            index: The index of the nonzero cotangent.
        Returns: A zeroed copy of x broadcasted to hold each of the cotangents, with up to one
            element holding x.
        """
        def f(x: jnp.ndarray) -> jnp.ndarray:
            z = jnp.zeros((len(self.cotangent_names),) + x.shape, dtype=x.dtype)
            if index is None:
                return z
            return z.at[index].set(x)

        return tree_map(f, x)



from __future__ import annotations

from collections.abc import Iterable
from functools import partial
from typing import (Any, Callable, Dict, Generic, Hashable, Mapping, Sequence, Tuple, Type, TypeVar,
                    cast)

from jax.tree_util import register_pytree_node_class

from .annotations import PyTree

__all__ = ['Partial']


R = TypeVar('R')


@register_pytree_node_class
class Partial(partial, Generic[R]):  # type: ignore
    """
    A version of functools.partial that returns a pytree.

    Use it for partial function evaluation in a way that is compatible with JAX's transformations,
    e.g., ``Partial(func, *args, **kwargs)``.

    JAX's version of this class in jax.tree_util assumes that all parameters are nonstatic and the
    callable is static.  This version allows you to specify which parameters are static and whether
    the callable is static.
    """

    T = TypeVar('T', bound='Partial[R]')

    callable_is_static: bool
    static_argnums: Tuple[int, ...]
    static_kwargs: Mapping[str, Any]

    # TODO: use positional-only arguments.
    def __new__(cls: Type[T],
                func: Callable[..., R],
                # /,
                *args: Any,
                callable_is_static: bool = True,
                static_argnums: Tuple[int, ...] = (),
                static_kwargs: Mapping[str, Any] = None,
                **kwargs: Any) -> T:
        """
        Args:
            func: The function being applied.
            args: The applied positional arguments.
            callable_is_static: Whether the function callable is static.
            static_argnums: The indices of the applied positional arguments that are static.
            static_kwargs: The key-value pairs representing applied keyword arguments that are
                static.
            kwargs: The applied keyword arguments.
        """
        if callable_is_static and isinstance(func, Partial):
            raise TypeError
        retval = super().__new__(cls, func, *args, **kwargs)  # type: ignore
        retval.callable_is_static = callable_is_static
        retval.static_argnums = set(static_argnums)
        retval.static_kwargs = {} if static_kwargs is None else static_kwargs
        return retval

    def tree_flatten(self: Partial[R]) -> Tuple[Sequence[PyTree], Hashable]:
        static_args = []
        tree_args = []

        def _append(is_static: bool, value: Any) -> None:
            if is_static:
                static_args.append(value)
            else:
                tree_args.append(value)

        _append(self.callable_is_static, self.func)
        for i, value in enumerate(self.args):
            _append(i in self.static_argnums, value)

        return ((list(reversed(tree_args)), self.keywords),
                (self.callable_is_static, self.static_argnums,
                 list(reversed(static_args)), self.static_kwargs))

    @classmethod
    def tree_unflatten(cls: Type[R],
                       static: Hashable,
                       trees: Sequence[PyTree]) -> Partial[R]:
        if not isinstance(static, Iterable):
            raise RuntimeError

        callable_is_static, static_argnums, static_args, static_kwargs = static

        if not isinstance(static_args, list):
            raise RuntimeError

        tree_args, tree_kwargs = trees

        if not isinstance(tree_args, list):
            raise RuntimeError
        if not isinstance(tree_kwargs, dict):
            raise RuntimeError

        tree_kwargs = cast(Dict[str, Any], tree_kwargs)

        args = []
        for i in range(len(static_args) + len(tree_args)):
            if i == 0:
                is_static = callable_is_static
            else:
                is_static = i - 1 in static_argnums
            if is_static:
                args.append(static_args.pop())
            else:
                args.append(tree_args.pop())

        return Partial[R](*args,
                          callable_is_static=callable_is_static,
                          static_argnums=static_argnums,
                          static_kwargs=static_kwargs,
                          **tree_kwargs)

    def __call__(self, *args: Any, **kwargs: Any) -> R:
        return super().__call__(*args, **self.static_kwargs, **kwargs)



from __future__ import annotations

from typing import Any, List, Optional, Tuple, Type, TypeVar, Union

import jax.numpy as jnp
import jax.random
import numpy as np

from .annotations import Array, BooleanArray, RealArray, RealNumeric, Shape, ShapeLike
from .dataclasses import dataclass

__all__ = ['Generator']


T = TypeVar('T', bound='Generator')


@dataclass
class Generator:
    """
    This class represents a JAX random number generator.  Unlike `numpy.Generator`, `tjax.Generator`
    has no mutating methods.  Instead, its generation methods return a new instance along with the
    generated tensor.
    """
    key: jnp.ndarray

    # Class methods --------------------------------------------------------------------------------
    @classmethod
    def from_seed(cls: Type[T], seed: int) -> T:
        return cls(jax.random.PRNGKey(seed))

    # New methods ----------------------------------------------------------------------------------
    def split(self, n: int = 2) -> List[Generator]:
        """
        Split a generator into a list of generators.
        """
        assert self.key.shape == (2,)
        keys = jax.random.split(self.key, n)
        return [Generator(key=key) for key in keys]

    def vmap_split(self, shape: ShapeLike) -> Generator:
        """
        Split a generator into a generator that can be passed to a vmapped function.
        """
        if isinstance(shape, int):
            shape = (shape,)
        else:
            shape = tuple(shape)
        prod_shape = int(np.prod(shape))
        keys = (self.key
                if prod_shape == 1
                else jax.random.split(self.key, prod_shape))
        # https://github.com/google/jax/issues/6473
        keys = keys.reshape(shape + (2,))  # type: ignore
        return Generator(keys)

    def bernoulli(self, p: RealNumeric, shape: Shape = ()) -> Tuple[BooleanArray, Generator]:
        new_g, this_g = self.split()
        return jax.random.bernoulli(this_g.key, p, shape), new_g

    def choice(self,
               a: Union[int, Array],
               shape: Shape = (),
               replace: bool = True,
               p: Optional[RealArray] = None) -> Tuple[Union[int, Array], Generator]:
        new_g, this_g = self.split()
        return jax.random.choice(this_g.key, a, shape, replace, p), new_g

    def gamma(self, gamma_shape: RealNumeric, shape: Shape = ()) -> Tuple[RealArray, Generator]:
        new_g, this_g = self.split()
        return jax.random.gamma(this_g.key, gamma_shape, shape), new_g

    def normal(self, std_dev: RealNumeric, shape: Shape = ()) -> Tuple[RealArray, Generator]:
        new_g, this_g = self.split()
        return std_dev * jax.random.normal(this_g.key, shape), new_g

    def uniform(self, minval: RealNumeric = 0.0, maxval: RealNumeric = 0.0, shape: Shape = ()) -> (
            Tuple[RealArray, Generator]):
        new_g, this_g = self.split()
        return jax.random.uniform(this_g.key, minval=minval, maxval=maxval, shape=shape), new_g

    # Magic methods --------------------------------------------------------------------------------
    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Generator):
            return NotImplemented
        return jnp.all(self.key == other.key)

    def __ne__(self, other: Any) -> bool:
        if not isinstance(other, Generator):
            return NotImplemented
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(tuple(np.ravel(self.key)))



from functools import reduce
from numbers import Number
from operator import add
from typing import Any, Collection, Optional

import jax.numpy as jnp
import numpy as np

from .annotations import ComplexArray, ComplexNumeric, RealNumeric, ShapeLike

__all__ = ['sum_tensors', 'is_scalar', 'abs_square', 'safe_divide']


def sum_tensors(tensors: Collection[ComplexArray],
                shape: Optional[ShapeLike] = None) -> ComplexArray:
    if not tensors:
        return jnp.zeros(shape)
    return reduce(add, tensors)


def is_scalar(x: Any) -> bool:
    return isinstance(x, Number) or isinstance(x, (np.ndarray, jnp.ndarray)) and x.shape == ()


def abs_square(x: ComplexNumeric) -> RealNumeric:
    return jnp.square(x.real) + jnp.square(x.imag)


def safe_divide(numerator: ComplexNumeric, denominator: RealNumeric) -> ComplexNumeric:
    return jnp.where(denominator > 0.0, numerator / denominator, jnp.inf)



from typing import Any, Callable, Generic, Tuple, TypeVar, Union, cast

import jax
from jax.tree_util import Partial

__all__ = ['jit', 'custom_jvp', 'custom_vjp']


R = TypeVar('R')
F = TypeVar('F', bound=Callable[..., Any])


def as_sorted_tuple(x: Union[int, Tuple[int, ...]]) -> Tuple[int, ...]:
    if isinstance(x, int):
        return (x,)
    return tuple(sorted(x))


def jit(func: F, **kwargs: Any) -> F:
    """
    This version of jit ensures that abstract methods stay abstract.
    """
    retval = cast(F, jax.jit(func, **kwargs))
    if hasattr(func, "__isabstractmethod__"):
        retval.__isabstractmethod__ = func.__isabstractmethod__  # type: ignore
    return retval


class custom_vjp(Generic[R]):
    """
    This is a shim class over jax.custom_vjp to:

    - allow custom_vjp to be used on methods, and
    - rename nondiff_argnums to static_argnums.
    """
    vjp: jax.custom_vjp

    def __init__(self,
                 fun: Callable[..., R],
                 static_argnums: Union[int, Tuple[int, ...]] = ()):
        """
        Args:
            fun: the function to decorate.
            static_argnums: The indices of the static arguments.
        """
        static_argnums = as_sorted_tuple(static_argnums)
        self.vjp = jax.custom_vjp(fun, nondiff_argnums=static_argnums)  # type: ignore

    def defvjp(self, fwd: Callable[..., Tuple[R, Any]], bwd: Callable[..., Any]) -> None:
        self.vjp.defvjp(fwd, bwd)  # type: ignore

    def __call__(self, *args: Any) -> R:
        return self.vjp(*args)

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)


class custom_jvp(Generic[R]):
    """
    This is a shim class over jax.custom_jvp to:

    - allow custom_vjp to be used on methods, and
    - rename nondiff_argnums to static_argnums.
    """
    def __init__(self,
                 fun: Callable[..., R],
                 static_argnums: Union[int, Tuple[int, ...]] = ()):
        """
        Args:
            fun: the function to decorate.
            static_argnums: The indices of the static arguments.
        """
        static_argnums = as_sorted_tuple(static_argnums)
        self.jvp = jax.custom_jvp(fun, nondiff_argnums=static_argnums)  # type: ignore

    def defjvp(self, jvp: Callable[..., R]) -> None:
        """
        Implement the custom forward pass of the custom derivative.

        Args:
            fwd: The custom forward pass.
        """
        self.jvp.defjvp(jvp)  # type: ignore

    def __call__(self, *args: Any) -> R:
        return self.jvp(*args)

    def __get__(self, instance: Any, owner: Any = None) -> Callable[..., R]:
        if instance is None:
            return self
        # Create a partial function application corresponding to a bound method.
        return Partial(self, instance)
